/*
 * Copyright (c) 2015-2020 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Aug 24 2015
 */
#include "acpom.h"

#include <sys/time.h>
#include <aio.h>
#include <err.h>
#include <unistd.h>

#include <cstdio>
#include <cmath>
#include <iostream>

#include <Eigen/Geometry>

#include "pom_c_types.h"

#include "codels.h"
#include "ukf.h"


/* --- Task filter ------------------------------------------------------ */

/** Codel pom_filter_start of task filter.
 *
 * Triggered by pom_start.
 * Yields to pom_exec.
 * Throws pom_e_sys.
 */
genom_event
pom_filter_start(pom_context_s **context, const pom_frame *frame,
                 const genom_context self)
{
  frame->open("robot", self);
  return pom_exec;
}


/** Codel pom_filter_exec of task filter.
 *
 * Triggered by pom_exec.
 * Yields to pom_pause_exec.
 * Throws pom_e_sys.
 */
genom_event
pom_filter_exec(pom_context_s **context,
                const pom_ids_offset_s *offset, pom_log_s **log_state,
                const pom_frame *frame, const genom_context self)
{
  static double avgperiod;
  static uint32_t slow;

  if (!(*context)) return pom_pause_exec;

  const pom_history_s &h = (*context)->history;
  pom_history_s::reverse_iterator todo, skip, i, j;
  struct timeval start, now;
  double date, mdate;

  /* measure start time */
  gettimeofday(&start, NULL);

  /* get the oldest non fused state */
  for(todo = h.rbegin(); todo != h.rend(); ++todo)
    if (!todo->fused) break;

  /* set filter state to the fused state before */
  if (todo == h.rend() || todo == h.rbegin()) {
    (*context)->filter.set((*context)->state, (*context)->cov);
    date = (*context)->ts.sec + 1e-9 * (*context)->ts.nsec;
  } else {
    todo--;
    (*context)->filter.set(todo->state, todo->cov);
    date = todo->measurement.ts.sec + 1e-9 * todo->measurement.ts.nsec;
    todo++;
  }

  /* process history */
  skip = h.rend();
  for(; todo != h.rend(); ++todo) {
    const pom_snapshot &s = *todo;
    s.fused = false;

    /* if we've been computing for more than the period, skip to the
     * last measurement to be fused - skipped items will be processed next
     * time */
    if (skip == h.rend()) {
      gettimeofday(&now, NULL);
      if ((now.tv_sec - start.tv_sec) * 1e3 +
          (now.tv_usec - start.tv_usec) * 1e-3 > pom_period_ms) {
        for(i = j = todo, j++; j != h.rend(); ++i, ++j) {
          if (i->fused && !j->fused) skip = i;
        }
        if (skip == h.rend()) {
          /* nothing can be skipped: continue as normal + disable the test */
          skip = h.rbegin();
        } else {
          todo = skip;
          (*context)->filter.set(todo->state, todo->cov);
          date = todo->measurement.ts.sec + 1e-9 * todo->measurement.ts.nsec;
          continue;
        }
      }
    }

    mdate = s.measurement.ts.sec + 1e-9 * s.measurement.ts.nsec;

    if (mdate < date) {
      /*
       * skip measurements older than current state - this should not happen,
       * except for the very first measurement in the component lifetime.
       */
      (*context)->filter.predict(0.);
      (*context)->filter.update(s.state, s.cov);
      (*context)->filter.set(s.state, s.cov);
      s.fused = true;
      continue;
    }

    (*context)->filter.predict(mdate - date);

    /* position */
    if (s.measurement.pos._present &&
        !std::isnan(s.measurement.pos._value.x)) {
      (*context)->filter.measure.p() <<
        s.measurement.pos._value.x,
        s.measurement.pos._value.y,
        s.measurement.pos._value.z;

      if (s.measurement.pos_cov._present) {
        const double *c = s.measurement.pos_cov._value.cov;

        (*context)->filter.measure.p_cov() <<
          c[0],  c[1],  c[3],
          c[1],  c[2],  c[4],
          c[3],  c[4],  c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (*context)->filter.measure.p_cov() <<
          1.,  0.,  0.,
          0.,  1.,  0.,
          0.,  0.,  1.;
      }
    }

    /* orientation */
    if (s.measurement.att._present && !s.measurement.intrinsic &&
        !std::isnan(s.measurement.att._value.qw)) {
      (*context)->filter.measure.q() <<
        s.measurement.att._value.qx,
        s.measurement.att._value.qy,
        s.measurement.att._value.qz,
        s.measurement.att._value.qw;

      if (s.measurement.att_cov._present) {
        const double *c = s.measurement.att_cov._value.cov;

        (*context)->filter.measure.q_cov() <<
          c[2], c[4], c[7], c[1],
          c[4], c[5], c[8], c[3],
          c[7], c[8], c[9], c[6],
          c[1], c[3], c[6], c[0];
      } else {
        /* no covariance provided, provide a pessimistic default */
        double qx = s.measurement.att._value.qx;
        double qy = s.measurement.att._value.qy;
        double qz = s.measurement.att._value.qz;
        double qw = s.measurement.att._value.qw;

        (*context)->filter.measure.q_cov() <<
          0.25*(1-qx*qx), -0.25*qx*qy,     -0.25*qx*qz,     -0.25*qw*qx,
          -0.25*qx*qy,     0.25*(1-qy*qy), -0.25*qy*qz,     -0.25*qw*qy,
          -0.25*qx*qz,    -0.25*qy*qz,      0.25*(1-qz*qz), -0.25*qw*qz,
          -0.25*qw*qx,    -0.25*qw*qy,     -0.25*qw*qz,      0.25*(1-qw*qw);
      }
    }

    /* magnetic field */
    if (s.measurement.att._present && s.measurement.intrinsic &&
        std::isnan(s.measurement.att._value.qw)) {
      (*context)->filter.measure.im() <<
        s.measurement.att._value.qx,
        s.measurement.att._value.qy,
        s.measurement.att._value.qz;

      if (s.measurement.att_cov._present) {
        const double *c = s.measurement.att_cov._value.cov;

        (*context)->filter.measure.im_cov() <<
          c[2], c[4], c[7],
          c[4], c[5], c[8],
          c[7], c[8], c[9];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (*context)->filter.measure.im_cov() <<
          1e-5, 0.,    0.,
          0.,    1e-5, 0.,
          0.,    0.,    1e-5;
      }
    }

    /* linear velocity */
    if (s.measurement.vel._present &&
        !std::isnan(s.measurement.vel._value.vx)) {
      (s.measurement.intrinsic ?
       (*context)->filter.measure.iv() :
       (*context)->filter.measure.v()) <<
        s.measurement.vel._value.vx,
        s.measurement.vel._value.vy,
        s.measurement.vel._value.vz;

      if (s.measurement.vel_cov._present) {
        const double *c = s.measurement.vel_cov._value.cov;

        (s.measurement.intrinsic ?
         (*context)->filter.measure.iv_cov() :
         (*context)->filter.measure.v_cov()) <<
          c[0], c[1], c[3],
          c[1], c[2], c[4],
          c[3], c[4], c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.intrinsic ?
         (*context)->filter.measure.iv_cov() :
         (*context)->filter.measure.v_cov()) <<
          1., 0., 0.,
          0., 1., 0.,
          0., 0., 1.;
      }
    }

    /* angular velocity */
    if (s.measurement.avel._present &&
        !std::isnan(s.measurement.avel._value.wx)) {
      (s.measurement.intrinsic ?
       (*context)->filter.measure.iw() :
       (*context)->filter.measure.w()) <<
        s.measurement.avel._value.wx,
        s.measurement.avel._value.wy,
        s.measurement.avel._value.wz;

      if (s.measurement.avel_cov._present) {
        const double *c = s.measurement.avel_cov._value.cov;

        (s.measurement.intrinsic ?
         (*context)->filter.measure.iw_cov() :
         (*context)->filter.measure.w_cov()) <<
          c[0], c[1], c[3],
          c[1], c[2], c[4],
          c[3], c[4], c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.intrinsic ?
         (*context)->filter.measure.iw_cov() :
         (*context)->filter.measure.w_cov()) <<
          0.25,  0.,    0.,
          0.,    0.25,  0.,
          0.,    0.,    0.25;
      }
    }

    /* linear acceleration - XXX when intrisic, assume accelerometer */
    if (s.measurement.acc._present &&
        !std::isnan(s.measurement.acc._value.ax)) {
      (s.measurement.intrinsic ?
       (*context)->filter.measure.iawg() :
       (*context)->filter.measure.a()) <<
        s.measurement.acc._value.ax,
        s.measurement.acc._value.ay,
        s.measurement.acc._value.az;

      if (s.measurement.acc_cov._present) {
        const double *c = s.measurement.acc_cov._value.cov;

        (s.measurement.intrinsic ?
         (*context)->filter.measure.iawg_cov() :
         (*context)->filter.measure.a_cov()) <<
          c[0], c[1], c[3],
          c[1], c[2], c[4],
          c[3], c[4], c[5];
      } else {
        /* no covariance provided, provide a pessimistic default */
        (s.measurement.intrinsic ?
         (*context)->filter.measure.iawg_cov() :
         (*context)->filter.measure.a_cov()) <<
          1., 0., 0.,
          0., 1., 0.,
          0., 0., 1.;
      }
    }

    (*context)->filter.update(s.state, s.cov);
    s.fused = true;

    /* update state for next iteration */
    (*context)->filter.set(s.state, s.cov);
    date = mdate;
  }

  /* current state */
  gettimeofday(&now, NULL);
  (*context)->filter.predict(now.tv_sec + 1e-6*now.tv_usec - date);
  (*context)->filter.update((*context)->state, (*context)->cov);
  (*context)->ts.sec = now.tv_sec;
  (*context)->ts.nsec = now.tv_usec * 1000;

  avgperiod =
    (1. - 0.4) * avgperiod +
    0.4 * (
      (now.tv_sec - start.tv_sec)*1e3 + (now.tv_usec - start.tv_usec)*1e-3)
    / pom_period_ms;
  if (avgperiod > 1.25) {
    slow++;
    if (slow % 500 == 2)
      warnx("slow computation (%fms) ...", avgperiod*pom_period_ms);
  } else if (slow) {
    if (slow > 1)
      warnx("recovered good computation speed after %d periods", slow);
    slow = 0;
  }

  /* publish */
  or_pose_estimator_state *r = frame->data("robot", self);
  if (r) {
    Eigen::Map<const Eigen::Matrix<double, 3, 1> > offp(offset->p);
    Eigen::Map<const Eigen::Quaternion<double> > offq(offset->q);

    r->ts = (*context)->ts;
    r->intrinsic = false;

    Eigen::Matrix<double, 3, 1> p =
      offq._transformVector((*context)->state.p()) + offp;
    r->pos._present = true;
    r->pos._value.x = p(0);
    r->pos._value.y = p(1);
    r->pos._value.z = p(2);
    Eigen::Quaternion<double> q =
      offq * Eigen::Quaternion<double>((*context)->state.q());
    r->att._present = true;
    r->att._value.qw = q.w();
    r->att._value.qx = q.x();
    r->att._value.qy = q.y();
    r->att._value.qz = q.z();
    {
      Eigen::Matrix<double, 7, 7> J;
      J.setZero();
      J.block<3, 3>(0, 0) = offq.matrix();
      J.block<3, 3>(3, 3) = offq.matrix();
      J(6, 6) = 1.;

      Eigen::Matrix<double, 7, 7> c =
        J * (*context)->state.pq_cov((*context)->cov) * J.transpose();

      r->pos_cov._present = true;
      for(uint8_t m = 0, k = 0; m < 3; m++)
        for(uint8_t n = 0; n <= m; n++, k++)
          r->pos_cov._value.cov[k] = c(m, n);

      double *qc = r->att_cov._value.cov;
      r->att_cov._present = true;
      qc[0] = c(6, 6);
      qc[1] = c(3, 6);  qc[2] = c(3, 3);
      qc[3] = c(4, 6);  qc[4] = c(4, 3);  qc[5] = c(4, 4);
      qc[6] = c(5, 6);  qc[7] = c(5, 3);  qc[8] = c(5, 4);  qc[9] = c(5, 5);

      double *pqc = r->att_pos_cov._value.cov;
      r->att_pos_cov._present = true;
      pqc[0] = c(6, 0);  pqc[1] = c(6, 1);  pqc[2] = c(6, 2);
      pqc[3] = c(3, 0);  pqc[4] = c(3, 1);  pqc[5] = c(3, 2);
      pqc[6] = c(4, 0);  pqc[7] = c(4, 1);  pqc[8] = c(4, 2);
      pqc[9] = c(5, 0); pqc[10] = c(5, 1); pqc[11] = c(5, 2);
    }

    Eigen::Matrix<double, 3, 1> v =
      offq._transformVector((*context)->state.v());
    r->vel._present = true;
    r->vel._value.vx = v(0);
    r->vel._value.vy = v(1);
    r->vel._value.vz = v(2);
    Eigen::Matrix<double, 3, 1> w =
      offq._transformVector((*context)->state.w());
    r->avel._present = true;
    r->avel._value.wx = w(0);
    r->avel._value.wy = w(1);
    r->avel._value.wz = w(2);
    {
      Eigen::Matrix<double, 6, 6> J;
      J.setZero();
      J.block<3, 3>(0, 0) = offq.matrix();
      J.block<3, 3>(3, 3) = offq.matrix();

      Eigen::Matrix<double, 6, 6> c =
        J * (*context)->state.vw_cov((*context)->cov) * J.transpose();

      r->vel_cov._present = true;
      for(uint8_t m = 0, k = 0; m < 3; m++)
        for(uint8_t n = 0; n <= m; n++, k++)
          r->vel_cov._value.cov[k] = c(m, n);

      r->avel_cov._present = true;
      for(uint8_t m = 3, k = 0; m < 6; m++)
        for(uint8_t n = 3; n <= m; n++, k++)
          r->avel_cov._value.cov[k] = c(m, n);
    }

    Eigen::Matrix<double, 3, 1> a =
      offq._transformVector((*context)->state.a());
    r->acc._present = true;
    r->acc._value.ax = a(0);
    r->acc._value.ay = a(1);
    r->acc._value.az = a(2);
    r->aacc._present = false;
    {
      Eigen::Matrix<double, 3, 3> J(offq.matrix());

      Eigen::Matrix<double, 3, 3> c =
        J * (*context)->state.a_cov((*context)->cov) * J.transpose();

      r->acc_cov._present = true;
      for(uint8_t m = 0, k = 0; m < 3; m++)
        for(uint8_t n = 0; n <= m; n++, k++)
          r->acc_cov._value.cov[k] = c(m, n);

      r->aacc_cov._present = false;
    }
  }
  frame->write("robot", self);

  /* log */
  if ((*log_state)->req.aio_fildes >= 0) {
    (*log_state)->total++;
    if ((*log_state)->total % (*log_state)->decimation == 0) {
      if ((*log_state)->pending) {
        if (aio_error(&(*log_state)->req) != EINPROGRESS) {
          (*log_state)->pending = false;
          if (aio_return(&(*log_state)->req) <= 0) {
            warn("log_state");
            close((*log_state)->req.aio_fildes);
            (*log_state)->req.aio_fildes = -1;
          }
        } else {
          (*log_state)->skipped = true;
          (*log_state)->missed++;
        }
      }
    }

    if ((*log_state)->req.aio_fildes >= 0 && !(*log_state)->pending) {
      double qw = r->att._value.qw;
      double qx = r->att._value.qx;
      double qy = r->att._value.qy;
      double qz = r->att._value.qz;
      double roll, pitch, yaw;

      roll  = atan2(2 * (qw*qx + qy*qz), 1 - 2 * (qx*qx + qy*qy));
      pitch = asin(2 * (qw*qy - qz*qx));
      yaw   = atan2(2 * (qw*qz + qx*qy), 1 - 2 * (qy*qy + qz*qz));

      (*log_state)->req.aio_nbytes = snprintf(
        (*log_state)->buffer, sizeof((*log_state)->buffer),
        "%s" pom_log_line_fmt " " pom_log_state_fmt "\n",
        (*log_state)->skipped ? "\n" : "",
        r->ts.sec, r->ts.nsec,
        r->intrinsic,
        r->pos._present, r->att._present,
        r->vel._present, r->avel._present,
        r->acc._present,

        r->pos._value.x, r->pos._value.y, r->pos._value.z,
        roll, pitch, yaw,
        r->vel._value.vx, r->vel._value.vy, r->vel._value.vz,
        r->avel._value.wx, r->avel._value.wy, r->avel._value.wz,
        r->acc._value.ax, r->acc._value.ay, r->acc._value.az,

        r->pos_cov._value.cov[0],
        r->pos_cov._value.cov[2],
        r->pos_cov._value.cov[5],
        r->att_cov._value.cov[2],
        r->att_cov._value.cov[5],
        r->att_cov._value.cov[9],
        r->vel_cov._value.cov[0],
        r->vel_cov._value.cov[2],
        r->vel_cov._value.cov[5],
        r->avel_cov._value.cov[0],
        r->avel_cov._value.cov[2],
        r->avel_cov._value.cov[5],
        r->acc_cov._value.cov[0],
        r->acc_cov._value.cov[2],
        r->acc_cov._value.cov[5]);

      if (aio_write(&(*log_state)->req)) {
        warn("log_state");
        close((*log_state)->req.aio_fildes);
        (*log_state)->req.aio_fildes = -1;
      } else
        (*log_state)->pending = true;

      (*log_state)->skipped = false;
    }
  }

  return pom_pause_exec;
}
