/*
 * Copyright (c) 2015-2020 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                      Anthony Mallet on Mon Aug 24 2015
 */
#include "acpom.h"

#include <sys/time.h>
#include <aio.h>
#include <err.h>
#include <unistd.h>

#include <cassert>
#include <cmath>
#include <cstdio>

#include "pom_c_types.h"
#include "codels.h"


/* --- Task io ---------------------------------------------------------- */


/** Codel pom_io_start of task io.
 *
 * Triggered by pom_start.
 * Yields to pom_read.
 * Throws pom_e_sys.
 */
genom_event
pom_io_start(pom_ids *ids, const genom_context self)
{
  struct timeval now;
  pom_snapshot s;

  ids->context = new pom_context_s;
  ids->history_length = 2; /* 2s by default */

  /* set initial state */
  gettimeofday(&now, NULL);

  ids->context->ts.sec = now.tv_sec;
  ids->context->ts.nsec = now.tv_usec * 1000;

  ids->context->state = ukf::state_s::Identity();
  ids->context->cov = ukf::state_s::dcov_s::Identity();

  ids->pmodel = ukf::state_s::pmodel;
  ids->max_jerk = ukf::state_s::max_dadt;
  ids->max_dw = ukf::state_s::max_dwdt;

  ids->magdir.x = ukf::measure_s::magdir(0);
  ids->magdir.y = ukf::measure_s::magdir(1);
  ids->magdir.z = ukf::measure_s::magdir(2);

  ids->offset.p[0] = ids->offset.p[1] = ids->offset.p[2] = 0.;
  ids->offset.q[0] = ids->offset.q[1] = ids->offset.q[2] = 0.;
  ids->offset.q[3] = 1.;

  /* init logging */
  try {
    ids->log_state = new pom_log_s;
    ids->log_measurements = new pom_log_s;
  } catch(...) {
    errno = ENOMEM;
    return pom_e_sys_error("init", self);
  }
  ids->log_state->req.aio_fildes = -1;
  ids->log_state->req.aio_offset = 0;
  ids->log_state->req.aio_buf = ids->log_state->buffer;
  ids->log_state->req.aio_nbytes = 0;
  ids->log_state->req.aio_reqprio = 0;
  ids->log_state->req.aio_sigevent.sigev_notify = SIGEV_NONE;
  ids->log_state->req.aio_lio_opcode = LIO_NOP;
  ids->log_state->pending = false;
  ids->log_state->skipped = false;

  ids->log_measurements->req.aio_fildes = -1;
  ids->log_measurements->req.aio_offset = 0;
  ids->log_measurements->req.aio_buf = ids->log_measurements->buffer;
  ids->log_measurements->req.aio_nbytes = 0;
  ids->log_measurements->req.aio_reqprio = 0;
  ids->log_measurements->req.aio_sigevent.sigev_notify = SIGEV_NONE;
  ids->log_measurements->req.aio_lio_opcode = LIO_NOP;
  ids->log_state->pending = false;
  ids->log_state->skipped = false;

  return pom_read;
}


/** Codel pom_io_read of task io.
 *
 * Triggered by pom_read.
 * Yields to pom_pause_read, pom_insert.
 * Throws pom_e_sys.
 */
genom_event
pom_io_read(const pom_measure *measure, sequence_pom_portinfo_s *ports,
            sequence_or_pose_estimator_state *measurements,
            const genom_context self)
{
  const or_pose_estimator_state *m;
  size_t i;

  for(i = 0; i < ports->_length; i++) {
    if (!ports->_buffer[i].name[0]) continue;

    /* get current port data */
    if (measure->read(ports->_buffer[i].name, self) != genom_ok) {
      ports->_buffer[i].name[0] = 0;
      continue;
    }
    m = measure->data(ports->_buffer[i].name, self);
    if (!m) continue;

    /* do not consider already read data */
    if (m->ts.nsec == ports->_buffer[i].last.nsec &&
        m->ts.sec == ports->_buffer[i].last.sec) continue;
    ports->_buffer[i].last = m->ts;

    /* record data */
    assert(measurements->_length < measurements->_maximum);
    measurements->_buffer[measurements->_length] = *m;
    measurements->_length++;
  }

  return measurements->_length ? pom_insert : pom_pause_read;
}


/** Codel pom_io_insert of task io.
 *
 * Triggered by pom_insert.
 * Yields to pom_pause_read.
 * Throws pom_e_sys.
 */
genom_event
pom_io_insert(sequence_or_pose_estimator_state *measurements,
              pom_context_s **context, double history_length,
              pom_log_s **log_measurements, const genom_context self)
{
  or_pose_estimator_state *m;
  pom_snapshot s;
  struct timeval now;
  ssize_t avail;
  double delay;
  size_t i;
  char *l;
  int n;

  gettimeofday(&now, NULL);

  /* check if logging is requested and possible */
  l = NULL;
  avail = 0;
  if ((*log_measurements)->req.aio_fildes >= 0) {
    if ((*log_measurements)->pending) {
      if (aio_error(&(*log_measurements)->req) != EINPROGRESS) {
        (*log_measurements)->pending = false;
        if (aio_return(&(*log_measurements)->req) <= 0) {
          warn("log_measurements");
          close((*log_measurements)->req.aio_fildes);
          (*log_measurements)->req.aio_fildes = -1;
        }
      } else {
        (*log_measurements)->skipped = true;
        (*log_measurements)->missed++;
      }
    }
    (*log_measurements)->total++;

    if ((*log_measurements)->req.aio_fildes >= 0 &&
        !(*log_measurements)->pending) {
      l = (*log_measurements)->buffer;
      avail = sizeof((*log_measurements)->buffer);

      if ((*log_measurements)->skipped) {
        n = snprintf(l, avail, "\n");
        l += n; avail -= n; assert(avail > 0 && "log buffer too small");
        (*log_measurements)->skipped = false;
      }
    }
  }

  /* insert new measurements with an empty state */
  s.fused = false;
  s.state = ukf::state_s::Identity();
  s.cov = ukf::state_s::dcov_s::Identity();

  for(i = 0; i < measurements->_length; i++) {
    m = &measurements->_buffer[i];
    s.measurement = *m;
    (*context)->history.insert(s);

    /* build log buffer */
    if (l) {
      double roll, pitch, yaw;

      if (!m->pos._present) {
        m->pos._value.x = m->pos._value.y = m->pos._value.z = 0.;
      }
      if (!m->att._present) {
        roll = pitch = yaw = 0.;
      } else {
        double qw = m->att._value.qw;
        double qx = m->att._value.qx;
        double qy = m->att._value.qy;
        double qz = m->att._value.qz;

        roll  = atan2(2 * (qw*qx + qy*qz), 1 - 2 * (qx*qx + qy*qy));
        pitch = asin(2 * (qw*qy - qz*qx));
        yaw   = atan2(2 * (qw*qz + qx*qy), 1 - 2 * (qy*qy + qz*qz));
      }

      if (!m->vel._present) {
        m->vel._value.vx = m->vel._value.vy = m->vel._value.vz = 0.;
      }
      if (!m->avel._present) {
        m->avel._value.wx = m->avel._value.wy = m->avel._value.wz = 0.;
      }

      if (!m->acc._present) {
        m->acc._value.ax = m->acc._value.ay = m->acc._value.az = 0.;
      }

      delay = now.tv_sec - m->ts.sec +
        1e-9 * (1e3 * now.tv_usec - m->ts.nsec);

      n = snprintf(
        l, avail, "%g " pom_log_line_fmt,
        delay, m->ts.sec, m->ts.nsec,
        m->intrinsic,
        m->pos._present, m->att._present,
        m->vel._present, m->avel._present,
        m->acc._present,
        m->pos._value.x, m->pos._value.y, m->pos._value.z,
        roll, pitch, yaw,
        m->vel._value.vx, m->vel._value.vy, m->vel._value.vz,
        m->avel._value.wx, m->avel._value.wy, m->avel._value.wz,
        m->acc._value.ax, m->acc._value.ay, m->acc._value.az);
      l += n; avail -= n; assert(avail > 0 && "log buffer too small");

#define pom_log_cov(m, x)                                               \
      do {                                                              \
        size_t j;                                                       \
                                                                        \
        for(j = 0;                                                      \
            j < sizeof((m)-> x ## _cov._value.cov)/sizeof(double);      \
            j++) {                                                      \
          double c = (m)-> x ## _cov._present ?                         \
            (m)-> x ## _cov._value.cov[j] : 0.;                         \
          n = snprintf(l, avail, pom_logfmt, c);                        \
          l += n; avail -= n;                                           \
          assert(avail > 0 && "log buffer too small");                  \
        }                                                               \
      } while(0)

      pom_log_cov(m, pos);
      pom_log_cov(m, att);
      pom_log_cov(m, att_pos);

      pom_log_cov(m, vel);
      pom_log_cov(m, avel);

      pom_log_cov(m, acc);

#undef pom_log_cov

      n = snprintf(l, avail, "\n");
      l += n; avail -= n; assert(avail > 0 && "log buffer too small");
    }
  }
  measurements->_length = 0;

  /* truncate history */
  s.measurement.ts.sec = now.tv_sec - (time_t)history_length;
  n = (history_length - (time_t)history_length)*1e6;
  if (now.tv_usec >= n)
    s.measurement.ts.nsec = (now.tv_usec - n) * 1000;
  else {
    s.measurement.ts.sec--;
    s.measurement.ts.nsec = (now.tv_usec + 1000000 - n) * 1000;
  }
  pom_history_s::iterator last = (*context)->history.upper_bound(s);

  if (last != (*context)->history.end()) {
    for(;!last->fused && last != (*context)->history.end(); ++last)
      warnx("history too short");

    if (last != (*context)->history.end())
      (*context)->history.erase(++last, (*context)->history.end());
  }

  /* issue log request */
  if (l) {
    (*log_measurements)->req.aio_nbytes =
      sizeof((*log_measurements)->buffer) - avail;
    if (aio_write(&(*log_measurements)->req)) {
      warn("log_measurements");
      close((*log_measurements)->req.aio_fildes);
      (*log_measurements)->req.aio_fildes = -1;
    } else
      (*log_measurements)->pending = true;
  }

  return pom_pause_read;
}


/** Codel pom_io_stop of task io.
 *
 * Triggered by pom_stop.
 * Yields to pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_io_stop(pom_ids *ids, const genom_context self)
{
  if (ids->context) delete ids->context;
  ids->context = NULL;

  pom_log_stop(&ids->log_state, &ids->log_measurements, self);
  delete ids->log_state;
  delete ids->log_measurements;
  ids->log_state = NULL;
  ids->log_measurements = NULL;

  return pom_ether;
}


/* --- Activity add_measurement ----------------------------------------- */

/** Codel pom_add_measurement of activity add_measurement.
 *
 * Triggered by pom_start.
 * Yields to pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_add_measurement(const char port[128],
                    sequence_pom_portinfo_s *ports,
                    sequence_or_pose_estimator_state *measurements,
                    const genom_context self)
{
  size_t i;

  for(i = 0; i < ports->_length; i++) {
    if (!strcmp(ports->_buffer[i].name, port)) return pom_ether;
  }

  if (i >= ports->_length) {
    if (genom_sequence_reserve(measurements, i + 1))
      return pom_e_sys_error("add", self);
    if (genom_sequence_reserve(ports, i + 1))
      return pom_e_sys_error("add", self);

    ports->_length = i + 1;
  }

  strncpy(
    ports->_buffer[i].name, port, sizeof(ports->_buffer->name));
  ports->_buffer[i].last.sec = 0;
  ports->_buffer[i].last.nsec = 0;

  return pom_ether;
}


/* --- Activity replay -------------------------------------------------- */

static FILE *replayf;
static double replayoff;

/** Codel pom_replay_start of activity replay.
 *
 * Triggered by pom_start.
 * Yields to pom_read, pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_replay_start(const char path[128], uint32_t *lineno,
                 const genom_context self)
{
  char header[1024];
  char *s;

  replayf = fopen(path, "r");
  if (!replayf) return pom_e_sys_error(path, self);

  /* discard header */
  s = fgets(header, sizeof(header), replayf);
  if (!s) return pom_stop;

  replayoff = nan("");
  *lineno = 2;
  return pom_read;
}

/** Codel pom_replay_read of activity replay.
 *
 * Triggered by pom_read.
 * Yields to pom_insert, pom_stop.
 * Throws pom_e_sys.
 */
genom_event
pom_replay_read(sequence_or_pose_estimator_state *measurements,
                uint32_t *lineno, const genom_context self)
{
  or_pose_estimator_state m;
  struct timeval now;
  struct timespec dt, rdt;
  int posp, attp, velp, avelp, accp;
  double t;
  int i, c;

  do {
    /* get next measurement */
    double roll, pitch, yaw;

    c = fscanf(replayf, "%*f " pom_scan_line_fmt,
               &m.ts.sec, &m.ts.nsec,
               &i, &posp, &attp, &velp, &avelp, &accp,
               &m.pos._value.x, &m.pos._value.y, &m.pos._value.z,
               &roll, &pitch, &yaw,
               &m.vel._value.vx, &m.vel._value.vy, &m.vel._value.vz,
               &m.avel._value.wx, &m.avel._value.wy, &m.avel._value.wz,
               &m.acc._value.ax, &m.acc._value.ay, &m.acc._value.az);
    if (c == EOF) return pom_stop;
    else if (c != 23) {
      warnx("bad state format, line %d", *lineno);
      errno = EBADF;
      return pom_e_sys_error("replay", self);
    }

    m.intrinsic = i;

    m.pos._present = m.pos_cov._present = posp;
    m.att._present = m.att_cov._present = attp;
    m.att_pos_cov._present = posp && attp;

    m.vel._present = m.vel_cov._present = velp;
    m.avel._present = m.avel_cov._present = avelp;

    m.acc._present = m.acc_cov._present = accp;
    m.aacc._present = m.aacc_cov._present = false;

    double cr = cos(roll/2.), sr = sin(roll/2.);
    double cp = cos(pitch/2.), sp = sin(pitch/2.);
    double cy = cos(yaw/2.), sy = sin(yaw/2.);

    m.att._value.qw = cy * cr * cp + sy * sr * sp;
    m.att._value.qx = cy * sr * cp - sy * cr * sp;
    m.att._value.qy = cy * cr * sp + sy * sr * cp;
    m.att._value.qz = sy * cr * cp - cy * sr * sp;

#define pom_scan_cov(f, m, x)                                           \
    do {                                                                \
      size_t j;                                                         \
                                                                        \
      for(j = 0;                                                        \
          j < sizeof((m). x ## _cov._value.cov)/sizeof(double);         \
          j++) {                                                        \
        if (!fscanf(f, "%lf", &(m). x ## _cov._value.cov[j])) {         \
          warnx("bad covariance format, line %d", *lineno);             \
          errno = EBADF;                                                \
          return pom_e_sys_error("replay", self);                       \
        }                                                               \
      }                                                                 \
    } while(0)

    pom_scan_cov(replayf, m, pos);
    pom_scan_cov(replayf, m, att);
    pom_scan_cov(replayf, m, att_pos);

    pom_scan_cov(replayf, m, vel);
    pom_scan_cov(replayf, m, avel);

    pom_scan_cov(replayf, m, acc);

#undef pom_scan_cov

    if (!fscanf(replayf, "\n"))
      /* tolerate missing newline, but consume it if present */;
    (*lineno)++;

    /* get time offset */
    gettimeofday(&now, NULL);
    if (std::isnan(replayoff)) {
      replayoff = now.tv_sec + 1e-6*now.tv_usec - m.ts.sec - 1e-9*m.ts.nsec;
    }

    t = m.ts.sec + 1e-9*m.ts.nsec + replayoff;
    m.ts.sec = floor(t);
    m.ts.nsec = 1e9*(t - m.ts.sec);

    /* get deadline */
    t = m.ts.sec + 1e-9*m.ts.nsec - now.tv_sec - 1e-6*now.tv_usec;

    /* record data */
    if (measurements->_length >= measurements->_maximum)
      if (genom_sequence_reserve(measurements, 1 + measurements->_maximum))
        return pom_e_sys_error("replay", self);

    measurements->_buffer[measurements->_length] = m;
    measurements->_length++;
  } while(t < -1e-3);

  /* wait until deadline */
  if (t > 1e-3) {
    dt.tv_sec = floor(t);
    dt.tv_nsec = 1e9*(t - dt.tv_sec);
    while(nanosleep(&dt, &rdt)) {
      if (errno != EINTR) break;
      dt = rdt;
    }
  }

  return pom_insert;
}

/** Codel pom_io_insert of activity replay.
 *
 * Triggered by pom_insert.
 * Yields to pom_pause_read.
 * Throws pom_e_sys.
 */
/* already defined in task io */


/** Codel pom_replay_stop of activity replay.
 *
 * Triggered by pom_stop.
 * Yields to pom_ether.
 * Throws pom_e_sys.
 */
genom_event
pom_replay_stop(const genom_context self)
{
  if (replayf) {
    fclose(replayf);
    replayf = NULL;
  }

  return pom_ether;
}
